use rand_chacha::ChaCha8Rng;
use rand_chacha::rand_core::SeedableRng;
use rand::seq::SliceRandom;

use wordlelike::constraints;
use wordlelike::constraints::Constraints;
use wordlelike::evaluator;
use wordlelike::evaluator::Evaluator;
use wordlelike::game;
use wordlelike::guess::*;
use wordlelike::guesser;

pub struct Puzzle<const N: usize> {
    pub history: GuessHistory<N>,
}

impl<const N: usize> Puzzle<N> {
    pub fn new(dictionary: &Dictionary<N>, seed: u64) -> Self {
        let mut rng: ChaCha8Rng = ChaCha8Rng::seed_from_u64(seed);
        let mut history: Option<GuessHistory<N>> = None;
        loop {
            let initial_secret = dictionary.choose(&mut rng).unwrap();
            let constraints = &mut constraints::Total::new();
            let mut game = game::Game {
                evaluator: &mut evaluator::SecretEvaluator::new(*initial_secret),
                guesser: &mut guesser::RandomConstraintGuesser::new(dictionary, constraints, &mut rng),
            };
            let mut game_history = Vec::new();
            loop {
                let result = game.play_round();
                if result.correct() { break; }
                game.push(result);
                game_history.push(result);
                if game_history.len() > 2 {
                    if result.matches
                        .iter()
                        .filter(|&m| *m == LetterMatch::HERE)
                        .count() < 3 {
                        continue;
                  }
                    let possible_words = game.guesser.get_possible_words();
                    if Puzzle::interesting(dictionary, &game_history, possible_words) {
                        history = Some(game_history);
                        break;
                    }
                }
            }
            if history.is_some() { break; }
        }
        Self {
            history: history.unwrap(),
        }
    }

    fn interesting(dictionary: &Dictionary<N>, history: &GuessHistory<N>, possible_words: Dictionary<N>) -> bool {
        let count = possible_words.len();
        if count < 3 || count > 6 {
            return false;
        }

        let base_constraints = &mut constraints::Total::new();
        let constraints = &mut constraints::Total::new();
        let mut evaluator = evaluator::AdversarialEvaluator::new(dictionary);
        for &result in history {
            base_constraints.push(result);
            evaluator.push(result);
        }

        let mut solutions = Vec::new();
        for &word in dictionary {
            constraints.clone_from(base_constraints);
            let matches = evaluator.evaluate(word);
            constraints.push(GuessResult { word, matches });
            let remaining_words = possible_words
                .iter()
                .filter(|&word| constraints.check(*word))
                .count();
            if remaining_words == 1 {
                solutions.push(word);
            }
        }

        let solution_count = solutions.len();
        let interesting = solution_count > 0 && solution_count <= 100;
        if interesting {
            println!("Could be: {:?}", possible_words);
        }
        interesting
    }
}